//
//  PLSession.h
//  Picturelife iOS SDK
//
//  Created by Glenn Wolters on 04/11/13.
//  Copyright (c) 2013 Glenn Wolters. All rights reserved.
//

#import <Foundation/Foundation.h>

/**
 * The PLSession class is an object that handles oauth with the Picturelife API. Hassle-free.
 *
 * An instance of PLSession can be used to manage a single user's access token and handles login, 
 * requesting and refreshing access tokens and keychain management. 
 *
 * The Picturelife authorization flow is a follows:
 * 1. Call authorizeWithSuccessBlock:failureBlock:
 * 2. Step two depends on the state of the session:
 * PLSessionStateInvalid:
 *  a. Redirect the user to Safari to login to Picturelife and authorize the application
 *  b. handleURL: is called from the UIApplicationDelegate containing a token (or code)
 *  c. A call is made to https://api.picturelife.com/oauth/access_token with the code, key, and secret
 *  d. The server response is stored in the keychain and the success block is called
 * PLSessionStateExpired
 *  a. Refresh the token by making a call to the Picturelife API containing the refresh token
 *  b. The server response is stored in the keychain and the success block is called
 * PLSessionStateActive
 *  a. Returns YES when an active PLSession was found in the keychain and the access token property of the PLSession object is set
 *
 * PLSession does not support access tokens for multiple user's, but this functionality could be added easily by subclasses.
 */
@class PLSession;

///-----------------------------
/// @name Enums
///-----------------------------
/**
 * The PLSessionState enumeration defines several states in which the session could be in during
 * the authorization process. When a method is called that does asynchronous request, that state is changed prior
 * to requests starting. The state of a session is PLSessionState by default.
 */
typedef NS_ENUM(NSUInteger, PLSessionState) {
    PLSessionStateUnknown = 0, // default state
    PLSessionStateRequestingCode, // requesting a token by switching to Safari
    PLSessionStateRequestingAccessToken, // using the token from Safari to get an access token
    PLSessionStateActive, // when the session has an active access token
    PLSessionStateExpired, // when the session's access token has expired, but before refreshing
    PLSessionStateRefreshingToken, // the session is refreshing its access token
    PLSessionStateInvalid // when the session was unable to parse a server response or when an error occured
};

typedef NS_ENUM(NSUInteger, PLSessionAuthorizationError) {
    PLSessionAuthorizationErrorUnknown = 0, // no specific information about the failure is known
    PLSessionAuthorizationErrorInvalidCode, // the code returned by a trip to Safari is invalid
    PLSessionAuthorizationErrorNetworkRequestFailed, // the access token or refresh token request failed
    PLSessionAuthorizationErrorInvalidResponse, // the response of the network request was invalid
    PLSessionAuthorizationErrorUnableToRefresh, // refreshing of the token failed
    PLSessionAuthorizationErrorUserCancelled // the user cancelled the token request and returned from Safari without a token
};

/**
 * The domain of errors that occured while authorizing with the Picturelife API
 */
extern NSString * const PLSessionErrorDomain;

///-----------------------------
/// @name Response Blocks
///-----------------------------
/**
 * The PLSessionAuthorizationSuccessBlock is passed to authorizeWithSuccessBlock:failureBlock and
 * returns a PLSession object when the authorization succeeded. The session object is the same as
 * the one that the authorizeWithSuccessBlock:failureBlock was called on.
 */
typedef void(^PLSessionAuthorizationSuccessBlock)(PLSession *session);

/**
 * The PLSessionAuthorizationFailureBlock returns when the authorization failed and has a
 * NSError object that describes the error that occurred. The possible error codes can be found
 * in the PLSessionAuthorizationError enumeration and the error domain is defined in PLSessionErrorDomain
 */
typedef void(^PLSessionAuthorizationFailureBlock)(NSError *error);

@interface PLSession : NSObject

///-----------------------------
/// @name Initialization
///-----------------------------
/**
 * Initializes a PLSession object with the given key and secret. The key and secret are used internally when making 
 * requests for tokens, access tokens or when refreshing an access token. 
 * 
 * @param appKey The key is called the client_id in Picturelife API requests, but key on the "register application" portal
 * @param appSecret The secret is called client_secret in the Picturelife API requests, but secret on the "register application" portal
 * @return an instance of PLSession, or subclass when called on a subclass of PLSession
 */
- (instancetype)initWithAppKey:(NSString *)appKey appSecret:(NSString *)appSecret;

///-----------------------------
/// @name Safari Redirect Handling
///-----------------------------
/**
 * When the UIApplicationDelegate gets a call to application:openURL:sourceApplication:annotation: it should be passed to this method.
 * It will check if the host of the url is in correct (see class documentation for info on URL schemes), and checks if the query
 * of the url contains a code that can be used to retrieve a access token. 
 * 
 * If this method is unable to handle the url authorizeWithSuccessBlock:failureBlock is called containing an NSError with the supplied url.
 * Will start a networking request to retrieve the access token asynchronously.
 *
 * @param url The url as passed to application:openURL:sourceApplication:annotation
 * @return whether or not the PLSession was able to handle the supplied url
 */
- (BOOL)handleURL:(NSURL *)url;

///-----------------------------
/// @name Request Accesstoken
///-----------------------------
/**
 * Will check to see if a session can be found in the keychain and returns YES when that session is (still). 
 * When an expired session was found a call to the Picturelife API is made requesting a refresh of that token, 
 * and when no session is found the user is switched to Safari to log in to the Picturelife website and authorize the appilication.
 *
 * The blocks are stored in a private instance variable and it is therefor not recommended to make several calls to
 * authorizeWithSuccessBlock:failureBlock: since this will override the previous blocks.
 *
 * @param successBlock The success block will be called when the session has successfuly authorized with the Picturelife API.
 * @param failureBlock The failure block will be called when the session failed to authorize with the Picturelife API because of one of the errors described in the PLSessionAuthorizationError enumeration.
 * @return will return YES when an active session was found in the keychain, returns NO when an asynchronous request for an access token or refreshing token is started
 */
- (BOOL)authorizeWithSuccessBlock:(PLSessionAuthorizationSuccessBlock)successBlock failureBlock:(PLSessionAuthorizationFailureBlock)failureBlock;

///-----------------------------
/// @name Expiration
///-----------------------------
/**
 * A helper function for checking if the session has expired. 
 * Checks the expire date against [NSDate date].
 *
 * @return whether the session has expired
 */
- (BOOL)isSessionExpired;

///-----------------------------
/// @name Session Properties
///-----------------------------
/**
 * @return The access token returned by the Picturelife API. Could be nil or invalid depending on the PLSessionState.
 */
@property (nonatomic, readonly) NSString *accessToken;

/**
 * @return The refresh token returned by the Picturelife API. Is used to refresh the access token when the session expires.
 */
@property (nonatomic, readonly) NSString *refreshToken;

/**
 * @return The expire date of the access token in UTC, based on a UNIX timestamp returned by the Picturelife API.
 */
@property (nonatomic, readonly) NSDate *expireDate;

/**
 * @return The user id to which the PLSession belongs. Use this to make personalized API calls or to identify the user in your application.
 */
@property (nonatomic, readonly) NSString *userID;

/**
 * @return The current state of the PLSession. The possible states and their explanation can be found in the PLSessionState enumeration.
 */
@property (nonatomic, readonly) PLSessionState state;

@end
