//
//  PLSession.m
//  Picturelife iOS SDK
//
//  Created by Glenn Wolters on 04/11/13.
//  Copyright (c) 2013 Glenn Wolters. All rights reserved.
//

#import "PLSession.h"
#import "PLHTTPRequestOperationManager.h"

// the account and service name for keychain management could be based on
// the app's key and secret, but this might make it hard to retrieve/delete old sessions
// when the key or secret changes
static NSString * const PLKeychainAccountName = @"PLKeychainAccount";
static NSString * const PLKeychainServiceName = @"PLKeychainService";

// the error domain used to create NSError objects containing descriptions
// of errors that occured during the authorization process.
NSString * const PLSessionErrorDomain = @"PLSessionDomain";

@interface PLSession ()

///-----------------------------
/// @name Safari Requests
///-----------------------------
/**
 * This method will compose a NSURL for requesting a token from Picturelife that is 
 * used to retrieve the actual access token.
 * The NSURL will be passed to Safari where it will request the user to login to its
 * Picturelife account and return to this application using the key that is also setup
 * as a url scheme in the info.plist
 */
- (void)requestOauthToken;

/**
 * Tries to handle the url that was passed to UIApplication's application:openURL:sourceApplication:annotation: method.
 * When the url's scheme is equal to the key used to init the session object, this method will make a
 * request to requestAccessTokenWithCode: using the code retrieved from the url.
 *
 * @param url The NSURL object that was passed to UIApplication's application:openURL:sourceApplication:annotation: method, should be of the following format [key]://code=[code]
 * @return returns whether or not the url was successfuly parsed and used to request an accesstoken from Picturelife.
 */
- (BOOL)handleURL:(NSURL *)url;

///-----------------------------
/// @name Networking Requests
///-----------------------------
/**
 * Starts a networking request to the Picturelife api using the PLHTTPRequestOperationManager.
 * The request will contain the key, secret that were used to init the session and the code 
 * that was returned from the Safari login. 
 * Changes the state of the session to PLSessionStateRequestingAccessToken.
 * @param code The code that was returned from a trip to Safari and its login to the Picturelife service
 */
- (void)requestAccessTokenWithCode:(NSString *)code;

/**
 * When the user has already authorized access to Picturelife before, but the session has expired, it then
 * uses this method the refresh the accesstoken using the refreshtoken that was also returned from the authorization request.
 * This method also starts a networking request using the PLHTTPRequestOperationManager, but adds a refresh_token and grant_type 
 * instead of the oauth token to the networking request.
 * Changes the state of the session to PLSessionStateRefreshingToken
 * @param session The previous session as retrieved from the keychain using +retrieveSession;
 */
- (void)refreshAccessTokenWithSession:(PLSession *)session;

///-----------------------------
/// @name State Management
///-----------------------------
/**
 * When the networking requests fail, or when the oauth token request in Safari fails, this method is called containing
 * an error description.
 * This changes the state of the session to PLSessionStateInvalid and dispatches the failure block used when calling
 * authorizeWithSuccess:failureBlock: async on the main queue. nil's the success and failure blocks afterwards
 * @param error An NSError object containing an error description of the error that occured during the authorization process
 */
- (void)authorizationRequestFailedWithError:(NSError *)error;

/**
 * When the networking request, for both refresh and access token, succeed this method is called with 
 * the repsonse of the networking request. It will first make a call to validateAuthorizationResponse: before
 * retrieving the access token, refresh token, expire date, and user id from the returned JSON object.
 * Changes the state of the session to PLSessionStateActive and stores the session in the keychain using the +storeSession: method.
 * Will call authorizationRequestFailedWithError: when the response is invalid
 * @param responseObject The response object returned by the API request. Should be a NSDictionary containing the server's JSON response.
 */
- (void)authorizationRequestSucceededWithResponseObject:(id)responseObject;

/**
 * Validates the authorization request's server response. Checks for a valid status code, whether the object is a NSDictionary
 * and whether the reponse contains a access token, refresh token and user id.
 * @param responseObject The response that returned by the server
 * @return returns whether or not the responseObject is a valid authorization response
 */
- (BOOL)validateAuthorizationResponse:(id)responseObject;

/**
 * Since state is a readonly property of the PLSession object, this method does not but change the instance variable
 * to the given state. Triggers no events nor does it send out notifications.
 */
- (void)setState:(PLSessionState)state;

///-----------------------------
/// @name Keychain Management
///-----------------------------
/**
 * This will store a session object in the keychain using PLKeychainAccountName and PLKeychainServiceName as parameters.
 * It will add an item to the keychain, if nil is returned by a call to +retrieveSession. If it returns a session object, 
 * this method will update the keychain item instead.
 * The session object is converted to NSData using NSCoding.
 * @param session The PLSession object that needs to be stored in the keychain as the most recent Picturelife session.
 * @return returns whether or not the keychain SecItemAdd or SecItemUpdate returned success.
 */
+ (BOOL)storeSession:(PLSession *)session;

/**
 * Retrieves a PLSession object from the keychain using PLKeychainAccountName and PLKeychainServiceName as parameters.
 * Because the PLSession object that was stored by +storeSession: was first converted to NSData, this method will convert
 * the NSData back to a PLSession object us NSCoding. 
 * If the keychain query does not return success then this method will return nil.
 * @return a PLSession object, if one was stored in the keychain before.
 */
+ (PLSession *)retrieveSession;

/**
 * Deletes the PLSession object from the keychain. Since the PLSession was stored without any specific knowledge of
 * the session itself, this method does nothing with the supplied PLSession object. But this object could be useful to
 * session objects that would like to subclass this method and delete the session based on user id or other identifiers.
 * @param session The session object that you would like to delete from the keychain (requires subclassing)
 * @return returns whether the keychain SecItemDeletet returned success
 */
+ (BOOL)deleteSession:(PLSession *)session;

@end

@implementation PLSession {
    NSString *_appKey;
    NSString *_appSecret;
    
    PLSessionAuthorizationSuccessBlock _authorizationSuccessBlock;
    PLSessionAuthorizationFailureBlock _authorizationFailureBlock;
}

- (instancetype)initWithAppKey:(NSString *)appKey appSecret:(NSString *)appSecret {
    self = [super init];
    
    if (self) {
        _appKey = appKey;
        _appSecret = appSecret;
    }
    
    return self;
}

- (BOOL)authorizeWithSuccessBlock:(PLSessionAuthorizationSuccessBlock)successBlock failureBlock:(PLSessionAuthorizationFailureBlock)failureBlock {

    // check if there is an active session stored in the keychain
    PLSession *activeSession = [[self class] retrieveSession];
    if (activeSession.state == PLSessionStateActive) {
        
        // assign the properties from the stored session to the current instance
        _accessToken = activeSession.accessToken;
        _refreshToken = activeSession.refreshToken;
        _expireDate = activeSession.expireDate;
        _userID = activeSession.userID;
        _state = PLSessionStateActive;
        
        return YES;
    }
    
    // maybe I should change these blocks to a property, for, you know, atomics
    _authorizationSuccessBlock = successBlock;
    _authorizationFailureBlock = failureBlock;
    
    // refresh the token if the state is expired
    if (activeSession.state == PLSessionStateExpired) {
        [self refreshAccessTokenWithSession:activeSession];
    }
    else {
        [self requestOauthToken];
    }
    
    return NO;
}

#pragma mark - Safari Request

- (void)requestOauthToken {
    [self setState:PLSessionStateRequestingCode];
    
    // this creates a formatted string containing the redirect url - the one that sends Safari back to this application
    // the url scheme should be the app key, as defined in the info.plist, followed by oauth.
    NSString *redirectURI = [NSString stringWithFormat:@"%@://oauth", _appKey];
    
    // this creates the url that sends to url to Safari, the Picturelife login screen and back to this application
    NSURL *tokenRequestURL = [NSURL URLWithString:[NSString stringWithFormat:@"https://api.picturelife.com/oauth/authorize?client_id=%@&client_secret=%@&redirect_uri=%@&response_type=code", _appKey, _appSecret, redirectURI]];
    
    [[UIApplication sharedApplication] openURL:tokenRequestURL];
}

- (BOOL)handleURL:(NSURL *)url {
    // only if the scheme is equal to the app key, do we know that this url was meant for us
    NSString *scheme = [url scheme];
    if ([_appKey isEqualToString:scheme] == NO) return NO;
    
    // the query should be composed of only the code=[this is the code]
    // this might not be fool proof, but works for the time being
    NSString *query = [url query];
    NSString *code = [query substringFromIndex:5];
    
    // fail the request with the url object in the userinfo if the code is nil
    if (code == nil) {
        NSDictionary *userInfo = [NSDictionary dictionaryWithObject:url forKey:@"url"];
        NSError *error = [NSError errorWithDomain:PLSessionErrorDomain
                                             code:PLSessionAuthorizationErrorInvalidCode
                                         userInfo:userInfo];
        [self authorizationRequestFailedWithError:error];
        return YES;
    }
    else {
        [self requestAccessTokenWithCode:code];
        return YES;
    }
}

#pragma mark - Authorization Requests

- (void)requestAccessTokenWithCode:(NSString *)code {
    [self setState:PLSessionStateRequestingAccessToken];
    
    // compose a dictionary with the query parameters
    // the key and secret are called client_id and client_secret in the API
    NSMutableDictionary *parameters = [[NSMutableDictionary alloc] init];
    [parameters setObject:_appKey forKey:@"client_id"];
    [parameters setObject:_appSecret forKey:@"client_secret"];
    [parameters setObject:code forKey:@"code"];
    
    // the networking request is done by the singleton PLHTTPRequestOperationManager that is a subclass of AFHTTPRequestOperation
    // initialized with the Picturelife API baseurl and possibly a secrity configuration
    // any other API requests can also be handled by this requestmanager from different classes
    PLHTTPRequestOperationManager *requestManager = [PLHTTPRequestOperationManager manager];
    [requestManager GET:@"oauth/access_token" parameters:parameters success:^(AFHTTPRequestOperation *operation, id responseObject) {
        
        [self authorizationRequestSucceededWithResponseObject:responseObject];
        
    } failure:^(AFHTTPRequestOperation *operation, NSError *responseError) {
        NSDictionary *userInfo = [NSDictionary dictionaryWithObject:responseError forKey:@"originalError"];
        NSError *error = [NSError errorWithDomain:PLSessionErrorDomain
                                             code:PLSessionAuthorizationErrorNetworkRequestFailed
                                         userInfo:userInfo];
        [self authorizationRequestFailedWithError:error];
    }];
}

- (void)refreshAccessTokenWithSession:(PLSession *)session {
    [self setState:PLSessionStateRefreshingToken];
    
    // get the refresh token from the session
    NSString *refreshToken = session.refreshToken;
    
    // the refresh token appears to be nil, lets get a new token instead
    if (refreshToken == nil) {
        [self requestOauthToken];
        return;
    }
    
    // the grant_type for refreshing should apparently be refresh_token :)
    NSMutableDictionary *parameters = [[NSMutableDictionary alloc] init];
    [parameters setObject:_appKey forKey:@"client_id"];
    [parameters setObject:_appSecret forKey:@"client_secret"];
    [parameters setObject:@"refresh_token" forKey:@"grant_type"];
    [parameters setObject:refreshToken forKey:@"refresh_token"];
    
    // you know what this does, don't you?
    PLHTTPRequestOperationManager *requestManager = [PLHTTPRequestOperationManager manager];
    [requestManager GET:@"oauth/token" parameters:parameters success:^(AFHTTPRequestOperation *operation, id responseObject) {
        
        [self authorizationRequestSucceededWithResponseObject:responseObject];
        
    } failure:^(AFHTTPRequestOperation *operation, NSError *responseError) {

        NSDictionary *userInfo = [NSDictionary dictionaryWithObject:responseError forKey:@"originalError"];
        NSError *error = [NSError errorWithDomain:PLSessionErrorDomain
                                             code:PLSessionAuthorizationErrorNetworkRequestFailed
                                         userInfo:userInfo];
        [self authorizationRequestFailedWithError:error];
    }];
}

#pragma mark - Authorization Completion

- (void)authorizationRequestFailedWithError:(NSError *)error {
    [self setState:PLSessionStateInvalid];
    
    dispatch_async(dispatch_get_main_queue(), ^{
        _authorizationFailureBlock(error);
        
        // I know, this might seem a bit tricky, since we're updating instance variables in a different queue
        // fix it if you have a better idea
        _authorizationSuccessBlock = nil;
        _authorizationFailureBlock = nil;
    });
}

- (void)authorizationRequestSucceededWithResponseObject:(id)responseObject {
    if ([self validateAuthorizationResponse:responseObject]) {
        
        // fail with invalid response
        NSError *error = [NSError errorWithDomain:PLSessionErrorDomain
                                             code:PLSessionAuthorizationErrorInvalidResponse
                                         userInfo:nil];
        
        [self authorizationRequestFailedWithError:error];
    }
    else {
        NSDictionary *responseDictionary = (NSDictionary *)responseObject;

        // convert the unix timestamp to an NSDate object for easier reference
        NSNumber *expire = [responseDictionary objectForKey:@"expires"];
        _expireDate = [NSDate dateWithTimeIntervalSince1970:[expire integerValue]];
        
        _accessToken = [responseDictionary objectForKey:@"access_token"];
        _refreshToken = [responseDictionary objectForKey:@"refresh_token"];
        _userID = [responseDictionary objectForKey:@"user_id"];
        
        [self setState:PLSessionStateActive];
        
        // store the session in the keychain
        [[self class] storeSession:self];
        
        dispatch_async(dispatch_get_main_queue(), ^{
            _authorizationSuccessBlock(self);
            
            // I know, this might seem a bit tricky, since we're updating instance variables in a different queue
            // fix it if you have a better idea
            _authorizationSuccessBlock = nil;
            _authorizationFailureBlock = nil;
        });
    }
}

- (BOOL)validateAuthorizationResponse:(id)responseObject {
    // only accept objects that are a NSDictionary
    if (responseObject == nil || [responseObject isKindOfClass:[NSDictionary class]] == NO) return NO;
    
    // check the status code of the API response
    // since I've been unable to get other responses from the Picturelife
    // I'm hoping that you could add support for other status codes when you encounter them
    NSDictionary *responseDictionary = (NSDictionary *)responseObject;
    NSNumber *statusCode = [responseDictionary objectForKey:@"status"];
    if ([statusCode integerValue] != 200) {
        NSLog(@"Invalid Status: %@ - Please add support for this status code if needed", statusCode);
        return NO;
    }
    
    // we need those keys baby!
    if ([responseDictionary objectForKey:@"access_token"] == nil ||
        [responseDictionary objectForKey:@"user_id"] == nil ||
        [responseDictionary objectForKey:@"expire"] == nil) {
        return NO;
    }
    
    return YES;
}

#pragma mark - Expiration

- (BOOL)isSessionExpired {
    NSComparisonResult compare = [self.expireDate isEqualToDate:[NSDate date]];
    return (compare == NSOrderedDescending);
}

#pragma mark - State

- (void)setState:(PLSessionState)state {
    _state = state;
}

#pragma mark - NSCoding

- (id)initWithCoder:(NSCoder *)decoder {
    self = [self init];

    if (self) {
        _accessToken = [decoder decodeObjectForKey:@"accessToken"];
        _refreshToken = [decoder decodeObjectForKey:@"refreshToken"];
        _expireDate = [decoder decodeObjectForKey:@"expireDate"];
        _userID = [decoder decodeObjectForKey:@"userID"];
        
        // since the session might've expired during its time in the keychain
        // we need to check if the session has an expire date and whether
        // the session is still active
        if (_expireDate != nil) {
            if ([self isSessionExpired] == NO) [self setState:PLSessionStateActive];
            else [self setState:PLSessionStateExpired];
        }
        else {
            [self setState:PLSessionStateInvalid];
        }
    }
    
    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder {
    // they see me codin' they hatin'
    [coder encodeObject:self.accessToken forKey:@"accessToken"];
    [coder encodeObject:self.refreshToken forKey:@"refreshToken"];
    [coder encodeObject:self.expireDate forKey:@"expireDate"];
    [coder encodeObject:self.userID forKey:@"userID"];
}

#pragma mark - Keychain

+ (BOOL)storeSession:(PLSession *)session {
    NSMutableDictionary *keychainQuery = [[NSMutableDictionary alloc] init];
    [keychainQuery setObject:(__bridge id)kSecClassGenericPassword forKey:(__bridge id)(kSecClass)];
    [keychainQuery setObject:PLKeychainAccountName forKey:(__bridge id)kSecAttrAccount];
    [keychainQuery setObject:PLKeychainServiceName forKey:(__bridge id)kSecAttrService];
    
    NSData *sessionData = [NSKeyedArchiver archivedDataWithRootObject:session];
    NSMutableDictionary *keychainItem = [[NSMutableDictionary alloc] init];
    [keychainItem setObject:sessionData forKey:(__bridge id)(kSecValueData)];
    
    // update the session if one already exists in the keychain
    BOOL sessionExists = ([self retrieveSession] != nil);
    if (sessionExists) {
        OSStatus updateStatus = SecItemUpdate((__bridge CFDictionaryRef)(keychainQuery), (__bridge CFDictionaryRef)(keychainItem));
        return (updateStatus == errSecSuccess);
    }
    else {
        [keychainItem addEntriesFromDictionary:keychainQuery];
        OSStatus addStatus = SecItemAdd((__bridge CFDictionaryRef)(keychainItem), NULL);
        return (addStatus == errSecSuccess);
    }
}

+ (PLSession *)retrieveSession {
    NSMutableDictionary *keychainQuery = [[NSMutableDictionary alloc] init];
    [keychainQuery setObject:(__bridge id)kCFBooleanTrue forKey:(__bridge id)(kSecReturnData)];
    [keychainQuery setObject:(__bridge id)kSecMatchLimitOne forKey:(__bridge id)kSecMatchLimit];
    [keychainQuery setObject:(__bridge id)kSecClassGenericPassword forKey:(__bridge id)(kSecClass)];
    [keychainQuery setObject:PLKeychainAccountName forKey:(__bridge id)kSecAttrAccount];
    [keychainQuery setObject:PLKeychainServiceName forKey:(__bridge id)kSecAttrService];
    
    CFDateRef queryResult = nil;
    OSStatus queryStatus = SecItemCopyMatching((__bridge CFDictionaryRef)(keychainQuery), (CFTypeRef *)&queryResult);
    
    if (queryStatus != errSecSuccess) {
        return nil;
    }
    
    // the query result is supposed to be an NSData object that can be converted to PLSession using NSCoding
    NSData *sessionData = (__bridge_transfer NSData *)queryResult;
    PLSession *session = [NSKeyedUnarchiver unarchiveObjectWithData:sessionData];
    
    return session;
}

+ (BOOL)deleteSession:(PLSession *)session {
    // creates a query for deleting the session from the keychain
    // does not keep in mind support for multiple session in the keychain
    NSMutableDictionary *keychainQuery = [[NSMutableDictionary alloc] init];
    [keychainQuery setObject:(__bridge id)kSecClassGenericPassword forKey:(__bridge id)(kSecClass)];
    [keychainQuery setObject:PLKeychainAccountName forKey:(__bridge id)kSecAttrAccount];
    [keychainQuery setObject:PLKeychainServiceName forKey:(__bridge id)kSecAttrService];
    
    OSStatus queryStatus = SecItemDelete((__bridge CFDictionaryRef)(keychainQuery));
    
    return (queryStatus == errSecSuccess);
}

@end
